""" outputs / abstractoutput.py """
import logging
from abc import ABC, abstractmethod

# from powermon.libs.mqttbroker import MqttBroker
from powermon.commands.result import Result
from powermon.dto.outputDTO import OutputDTO
from powermon.outputformats.abstractformat import AbstractFormat

log = logging.getLogger("Output")


class AbstractOutput(ABC):
    """ base class for all output modules """
    def __init__(self, name=None) -> None:
        self.name = name
        # self.command_code : str = "not_set"
        # self.device_id : str = "not_set"
        self.topic = None

    @property
    def formatter(self):
        """ the formatter for this output """
        return getattr(self, "_formatter", None)

    @formatter.setter
    def formatter(self, formatter : AbstractFormat):
        self._formatter = formatter

    @abstractmethod
    def process(self, command=None, result: Result = None, mqtt_broker=None, device_info=None):
        """ entry point of any output class """
        raise NotImplementedError("need to implement process function")

    def to_dto(self) -> OutputDTO:
        """ convert output object to a data transfer object """
        if self.formatter is None:
            format_dto = "None"
        else:
            format_dto = self.formatter.to_dto()
        return OutputDTO(type=self.name, format=format_dto)



""" formats / abstractformat.py """
import logging
import re
from abc import ABC, abstractmethod

from powermon.commands.reading import Reading
from powermon.commands.result import Result
# from powermon.device import DeviceInfo
from powermon.dto.formatDTO import FormatDTO

log = logging.getLogger("Formatter")


def pad(text, length):
    """ expand supplied value to 'length' with spaces, return unchanged if already longer """
    if isinstance(text, float) or isinstance(text, int):
        text = str(text)
    if len(text) > length:
        return text
    return text.ljust(length, " ")


def get_max_response_lengths(readings: list[Reading]):
    """ find the max length from a list of Readings """
    _max_name = 0
    _max_value = 0
    _max_unit = 0
    for reading in readings:
        _max_name = max(len(reading.data_name), _max_name)
        _max_value = max(len(str(reading.data_value)), _max_value)
        _max_unit = max(len(reading.data_unit), _max_unit)
    return _max_name, _max_value, _max_unit


class AbstractFormat(ABC):
    """ base for all format types """
    def __str__(self):
        return f"Format: {self.name}"

    def __init__(self, config=None):
        self.name = "AbstractFormat"
        if config is None:
            config = {}
        self.remove_spaces = config.get("remove_spaces", True)
        self.keep_case = config.get("keep_case", False)
        self.key_filter = config.get("filter", None)
        self.key_exclusion_filter = config.get("excl_filter", None)
        self.extra_info = config.get("extra_info", False)

    @property
    def key_filter(self):
        """ regular expression use to find keys wanted - or None to keep all """
        return getattr(self, "_key_filter", None)

    @key_filter.setter
    def key_filter(self, filter_string):
        if filter_string is not None:
            self._key_filter = re.compile(filter_string)

    @property
    def key_exclusion_filter(self):
        """ regular expression use to find keys to be excluded - or None to exclude none """
        return getattr(self, "_key_exclusion_filter", None)

    @key_exclusion_filter.setter
    def key_exclusion_filter(self, filter_string):
        if filter_string is not None:
            self._key_exclusion_filter = re.compile(filter_string)

    @abstractmethod
    def format(self, command, result: Result, device_info) -> list:
        """ entry point for all formats """

    def to_dto(self) -> FormatDTO:
        return FormatDTO(type=self.name)

    def format_and_filter_data(self, result: Result) -> list[Reading]:
        """ reformat key and remove unwanted readings """
        display_data = []
        for reading in result.readings:
            if reading is None:
                raise ValueError("reading cannot be None")
            formatted_key = self.format_key(reading.data_name)
            if self.is_key_wanted(formatted_key):
                display_data.append(reading)
        return display_data

    def format_key(self, key) -> str:
        """ reformat key """
        if self.remove_spaces:
            key = key.replace(" ", "_")
        if not self.keep_case:
            key = key.lower()
        return key

    def is_key_wanted(self, key) -> bool:
        """ determine if this item is wanted """
        # remove any specifically excluded keys
        if self.key_exclusion_filter is not None and self.key_exclusion_filter.search(key):
            # log.debug(f"key_wanted: key {key} matches excl_filter {excl_filter} so key excluded")
            return False
        if self.key_filter is None:
            # log.debug(
            #    f"key_wanted: No filter and key {key} not excluded by excl_filter {excl_filter} so key wanted"
            # )
            return True
        elif self.key_filter.search(key):
            # log.debug(
            #    f"key_wanted: key {key} matches filter {filter} and not excl_filter {excl_filter} so key wanted"
            # )
            return True
        else:
            # log.debug(f"key_wanted: key {key} does not match filter {filter} so key excluded")
            return False



""" formats / table.py """
import logging

from powermon.commands.reading import Reading
from powermon.commands.reading_definition import ReadingDefinition
from powermon.commands.result import Result
from powermon.errors import ConfigError
from powermon.outputformats.abstractformat import AbstractFormat

log = logging.getLogger("Table")


class Table(AbstractFormat):
    """ table formatter - formats results in a table suitable for std out """

    def __str__(self):
        return "outputs the results to standard out in a table (optionally formatted with line art boxes)"

    def __init__(self, config):
        super().__init__(config)
        self.name = "table"
        self.draw_lines = config.get("draw_lines", False)
        # self.command_description = "unknown command"

    # def set_command_description(self, command_description):
    #     self.command_description = command_description

    def format(self, command, result: Result, device_info) -> list[str]:
        log.info("Using output formatter: %s", self.name)

        _result = []
        filtered_responses: list[Reading] = []

        # check for error in result
        if result.error:
            filtered_responses.append(Reading(raw_value=None, processed_value=len(result.error_messages), definition=ReadingDefinition.from_config({"description": "Error Count"})))
            for i, message in enumerate(result.error_messages):
                reading_definition = ReadingDefinition.from_config({"description": f"Error #{i}"})
                filtered_responses.append(Reading(raw_value=None, processed_value=f"{message}", definition=reading_definition))

        if len(result.readings) == 0:
            filtered_responses.append(Reading(raw_value=None, processed_value="No readings in result", definition=ReadingDefinition.from_config({"description": "Error"})))

        filtered_responses.extend(self.format_and_filter_data(result))
        log.debug("displayData: %s", "\n".join((str(a) for a in filtered_responses)))

        # Determine column widths
        _pad = 1

        # data_name column
        max_data_name_length = max(len(reading.data_name) + _pad for reading in filtered_responses)
        max_data_name_length = max(max_data_name_length, 9 + _pad)
        # data_valuye column
        max_data_value_length = max(len(str(reading.data_value)) + _pad for reading in filtered_responses)
        max_data_value_length = max(max_data_value_length, 5 + _pad)
        # data_unit column
        max_data_unit_length = max(len(reading.data_unit) + _pad for reading in filtered_responses)
        max_data_unit_length = max(max_data_unit_length, 4 + _pad)

        # Total line length - of data
        line_length = max_data_name_length + max_data_value_length + max_data_unit_length + _pad

        cmd_str = f"Command: {command.code} - {command.command_definition.description}"

        if not self.extra_info:
            if self.draw_lines:
                # draw lines
                line_length += 3
                # Make line length longer if cmd_str exceeds data size
                line_length = max(line_length, len(cmd_str) + 4)
                # Check if columns too short and expand units if needed
                if (max_data_name_length + max_data_value_length + max_data_unit_length + 7) <= line_length:
                    max_data_unit_length = line_length - (max_data_name_length + max_data_value_length + 7)
                # Check if data and parameter are larger than line_length
                column_label_line_length = max(max_data_name_length, 9) + max(max_data_value_length, 5) + max(max_data_unit_length, 4) + 7
                line_length = max(column_label_line_length, line_length)
                # Command head / description
                _result.append(f"{'╔':═<{line_length - 1}}╗")
                _result.append(f"║ {cmd_str: <{line_length - 3}}║")
                _result.append(f"{'╠':═<{max_data_name_length + 2}}{'╤':═<{max_data_value_length + 2}}{'╤':═<{max_data_unit_length + 2}}╣")
                # Column titles
                _result.append(f"║ {'Parameter': <{max_data_name_length}}│ {'Value': <{max_data_value_length}}│ {'Unit': <{max_data_unit_length}}║")
                _result.append(f"{'╟':─<{max_data_name_length + 2}}{'┼':─<{max_data_value_length + 2}}{'┼':─<{max_data_unit_length + 2}}╢")
                # Data
                for reading in filtered_responses:
                    name = self.format_key(reading.data_name)
                    value = reading.data_value
                    unit = reading.data_unit
                    _result.append(f"║ {name: <{max_data_name_length}}│ {value: <{max_data_value_length}}│ {unit: <{max_data_unit_length}}║")
                # Footer
                _result.append(f"{'╚':═<{max_data_name_length + 2}}{'╧':═<{max_data_value_length + 2}}{'╧':═<{max_data_unit_length + 2}}╝")
            else:  # no lines
                # Make line length longer if cmd_str exceeds data size
                line_length = max(line_length, len(cmd_str) + _pad)
                # Command head / description
                _result.append(f"{'':-<{line_length}}")
                _result.append(f"{cmd_str}")
                _result.append(f"{'':-<{line_length}}")
                # Column titles
                _result.append(f"{'Parameter': <{max_data_name_length}} {'Value': <{max_data_value_length}} {'Unit': <{max_data_unit_length}}")
                # Data
                for reading in filtered_responses:
                    name = self.format_key(reading.data_name)
                    value = reading.data_value
                    unit = reading.data_unit
                    _result.append(f"{name: <{max_data_name_length}} {value: <{max_data_value_length}} {unit: <{max_data_unit_length}}")
        else:  # extra_info
            if self.draw_lines:
                raise ConfigError('drawlines with extra info is not supported')
            else:  # no lines
                # max_data_unit_length = max(len(reading.data_unit) + _pad for reading in filtered_responses)
                # Make line length longer if cmd_str exceeds data size
                line_length = max(line_length, len(cmd_str) + _pad)
                # Command head / description
                _result.append(f"{'':-<{line_length}}")
                _result.append(f"{cmd_str}")
                _result.append(f"{'':-<{line_length}}")
                # Column titles
                _result.append(f"{'Parameter': <{max_data_name_length}} {'Value': <{max_data_value_length}} {'Unit': <{max_data_unit_length}} Extra Info")
                # Data
                for reading in filtered_responses:
                    name = self.format_key(reading.data_name)
                    value = reading.data_value
                    unit = reading.data_unit
                    icon = reading.icon
                    state_class = reading.state_class
                    device_class = reading.device_class
                    extra_info = ''
                    if icon:
                        extra_info += f"icon:{icon}, "
                    if state_class:
                        extra_info += f"state_class:{state_class}, "
                    if device_class:
                        extra_info += f"device_class:{device_class}"
                    _result.append(f"{name: <{max_data_name_length}} {value: <{max_data_value_length}} {unit: <{max_data_unit_length}} {extra_info}")

        return _result



import logging
from powermon.outputformats.abstractformat import AbstractFormat
from powermon.commands.result import Result

log = logging.getLogger("raw")


class raw(AbstractFormat):
    """ return the raw response """
    def __init__(self, config):
        super().__init__(config)
        self.name = "raw"

    def format(self, command, result: Result, device_info):
        log.info("Using output formatter: %s", self.name)

        data = result.raw_response
        log.debug("raw formatter data: %s", data)
        return [data]



""" powermon / libs / mqttbroker.py """
import logging
from time import sleep

import paho.mqtt.client as mqtt_client

# Set-up logger
log = logging.getLogger("mqttbroker")


class MqttBroker:
    """ Wrapper for mqtt broker connectivity and message proccessing """
    def __str__(self):
        if self.disabled:
            return "MqttBroker DISABLED"
        else:
            return f"MqttBroker name: {self.name}, port: {self.port}, user: {self.username}"

    @classmethod
    def from_config(cls, config=None) -> 'MqttBroker':
        """ build the mqtt broker object from a config dict """
        log.debug("mqttbroker config: %s", config)

        if config:
            name = config.get("name")
            port = config.get("port", 1883)
            username = config.get("username")
            password = config.get("password")
            mqtt_broker = cls(name=name, port=port, username=username, password=password)
            mqtt_broker.adhoc_topic = config.get("adhoc_topic")
            return mqtt_broker
        else:
            return cls(name=None)

    def __init__(self, name, port=None, username=None, password=None):
        self.name = name
        self.port = port
        self.username = username
        self.password = password
        self.is_connected = False

        if self.name is None:
            self.disabled = True
        else:
            self.disabled = False
            self.mqttc = mqtt_client.Client()

    def on_connect(self, client, userdata, flags, rc):
        """ callback for connect """
        # 0: Connection successful
        # 1: Connection refused - incorrect protocol version
        # 2: Connection refused - invalid client identifier
        # 3: Connection refused - server unavailable
        # 4: Connection refused - bad username or password
        # 5: Connection refused - not authorised
        # 6-255: Currently unused.
        log.debug("on_connect called - client: %s, userdata: %s, flags: %s", client, userdata, flags)
        connection_result = [
            "Connection successful",
            "Connection refused - incorrect protocol version",
            "Connection refused - invalid client identifier",
            "Connection refused - server unavailable",
            "Connection refused - bad username or password",
            "Connection refused - not authorised",
        ]
        log.debug("MqttBroker connection returned result: %s %s", rc, connection_result[rc])
        if rc == 0:
            self.is_connected = True
            return
        self.is_connected = False

    def on_disconnect(self, client, userdata, rc):
        """ callback for disconnect """
        log.debug("on_disconnect called - client: %s, userdata: %s, rc: %s", client, userdata, rc)
        self.is_connected = False

    def connect(self):
        """ connect to mqtt broker """
        if self.disabled:
            log.info("MQTT broker not enabled, was a broker name defined? '%s'", self.name)
            return
        if not self.name:
            log.info("MQTT could not connect as no broker name")
            return
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect
        # if name is screen just return without connecting
        if self.name == "screen":
            # allows checking of message formats
            return
        try:
            log.debug("Connecting to %s on port %s", self.name, self.port)
            if self.username:
                # auth = {"username": mqtt_user, "password": mqtt_pass}
                _password = "********" if self.password is not None else "None"
                log.info("Using mqtt authentication, username: %s, password: %s", self.username, _password)
                self.mqttc.username_pw_set(self.username, password=self.password)
            else:
                log.debug("No mqtt authentication used")
                # auth = None
            self.mqttc.connect(self.name, port=self.port, keepalive=60)
            self.mqttc.loop_start()
            sleep(1)
        except ConnectionRefusedError as ex:
            log.warning("%s refused connection with error: '%s'", self.name, ex)

    def start(self):
        """ start mqtt broker """
        if self.disabled:
            return
        if self.is_connected:
            self.mqttc.loop_start()

    def stop(self):
        """ stop mqtt broker """
        log.debug("Stopping mqttbroker connection")
        if self.disabled:
            return
        self.mqttc.loop_stop()

    # def set(self, variable, value):
    #     setattr(self, variable, value)

    # def update(self, variable, value):
    #     # only override if value is not None
    #     if value is None:
    #         return
    #     setattr(self, variable, value)

    def subscribe(self, topic, callback):
        """ subscribe to a mqtt topic """
        if not self.name:
            return
        if self.disabled:
            return
        # check if connected, connect if not
        if not self.is_connected:
            log.debug("Not connected, connecting")
            self.connect()
        # Register callback
        self.mqttc.on_message = callback
        if self.is_connected:
            # Subscribe to command topic
            log.debug("Subscribing to topic: %s", topic)
            self.mqttc.subscribe(topic, qos=0)
        else:
            log.warning("Did not subscribe to topic: %s as not connected to broker", topic)

    def publish(self, topic: str = None, payload: str = None):
        """ function to publish messages to mqtt broker """
        if self.disabled:
            log.debug("Cannot publish msg as mqttbroker disabled")
            return
        # log.debug("Publishing '%s' to '%s'", payload, topic)
        if self.name == "screen":
            print(f"mqtt debug - topic: '{topic}', payload: '{payload}'")
            return
        # check if connected, connect if not
        if not self.is_connected:
            log.debug("Not connected, connecting")
            self.connect()
            sleep(1)
            if not self.is_connected:
                log.warning("mqtt broker did not connect")
                return
        if isinstance(topic, bytes):
            topic = topic.decode('utf-8')
        if isinstance(payload, bytes):
            payload = payload.decode('utf-8')
        try:
            infot = self.mqttc.publish(topic, payload)
            infot.wait_for_publish(5)
        except Exception as e:
            log.warning(str(e))

    @property
    def adhoc_topic(self) -> str:
        """ return the adhoc command topic """
        return getattr(self, "_adhoc_topic", None)

    @adhoc_topic.setter
    def adhoc_topic(self, value):
        log.debug("setting adhoc topic to: %s", value)
        self._adhoc_topic = value

    # def setAdhocCommands(self, config={}, callback=None):
    #     if not config:
    #         return
    #     if self.disabled:
    #         log.debug("Cannot setAdhocCommands as mqttbroker disabled")
    #         return

    #     adhoc_commands = config.get("adhoc_commands")
    #     # sub to command topic if defined
    #     adhoc_commands_topic = adhoc_commands.get("topic")
    #     if adhoc_commands_topic is not None:
    #         log.info("Setting adhoc commands topic to %s", adhoc_commands_topic)
    #         self.subscribe(adhoc_commands_topic, callback)



""" formats / json_fmt.py """
import json
import logging
from powermon.outputformats.abstractformat import AbstractFormat
from powermon.dto.formatDTO import FormatDTO
from powermon.commands.result import Result
from powermon.commands.reading import Reading

log = logging.getLogger("json")


class Json(AbstractFormat):
    """ simple format - {name}={value}{unit} format """
    def __init__(self, config):
        super().__init__(config)
        self.name = "json"
        self.json_format = config.get("format", "basic")
        self.include_missing = config.get("include_missing", False)

    # def set_command_description(self, command_description):
    #     pass

    def format(self, command, result: Result, device_info) -> list:

        _result = []
        # if jsonpickle is None:
        #     return _result

        # check for error in result
        if result.error:
            _result.append(f"Error Count: {len(result.error_messages)}")
            for i, message in enumerate(result.error_messages):
                _result.append(f"Error #{i}: {message}")
            # return _result

        if len(result.readings) == 0:
            return _result

        display_data : list[Reading] = self.format_and_filter_data(result)

        # build data to display
        for reading in display_data:
            # match self.json_format:
            #     case 'basic':  # only basic format type so far
            pickle = {'data_name': self.format_key(reading.data_name)}
            wanted_items = ['data_value', 'data_unit']
            if self.extra_info:
                wanted_items.extend(['icon', 'state_class', 'device_class'])

            if self.include_missing:
                pickle.update({key: getattr(reading, key) for key in wanted_items})
            else:
                pickle.update({key: getattr(reading, key) for key in wanted_items if getattr(reading, key) is not None})
            _result.append(json.dumps(pickle))
        return _result

    @classmethod
    def from_dto(cls, dto: FormatDTO):
        """ build class object from dto """
        return cls(config=dto)



""" outputformats / __init__.py """
import logging
from enum import Enum, auto

# from strenum import LowercaseStrEnum

from powermon.outputformats.abstractformat import AbstractFormat

log = logging.getLogger("formats")


class FormatterType(Enum):
    """ enumeration of valid formatter types """
    HASS = auto()
    HTMLTABLE = auto()
    JSON = auto()
    RAW = auto()
    SIMPLE = auto()
    TABLE = auto()
    TOPICS = auto()


DEFAULT_FORMAT = FormatterType.SIMPLE


def from_config(format_config) -> AbstractFormat:
    """ use a config dict to build and return a format class """
    # Get values from config
    log.debug("Format from_config, format_config: %s", format_config)

    # formatConfig can be None, a str (eg 'simple') or a dict
    if format_config is None:
        format_type = FormatterType.SIMPLE
        format_config = {}
    elif isinstance(format_config, str):
        format_type = format_config
        format_config = {}
        format_config["type"] = format_type
    else:
        format_type = format_config.get("type")
    log.debug("getFormatfromConfig, formatType: %s", format_type)

    match format_type:
        case FormatterType.HTMLTABLE:
            from powermon.outputformats.htmltable import HtmlTable as fmt
        case FormatterType.HASS:
            from powermon.outputformats.hass import Hass as fmt
        case FormatterType.JSON:
            from powermon.outputformats.json_fmt import Json as fmt
        # case FormatterType.TOPICS:
        #     from powermon.outputformats.topics import Topics as fmt
        case FormatterType.SIMPLE:
            from powermon.outputformats.simple import SimpleFormat as fmt
        case FormatterType.TABLE:
            from powermon.outputformats.table import Table as fmt
        case FormatterType.RAW:
            from powermon.outputformats.raw import raw as fmt
        case _:
            log.warning("No formatter found for: %s", format_type)
            return None
    return fmt(format_config)



"""
mqtt output module
outputs messages to mqtt broker
"""
import logging

from powermon.commands.result import Result
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("MQTT")


class MQTT(AbstractOutput):
    """ mqtt output class"""
    def __init__(self, topic: str):
        super().__init__(name="Mqtt")
        self.topic = topic

    def __str__(self):
        return f"outputs.MQTT: {self.topic=}"

    @property
    def topic(self):
        """ the topic to send the output to """
        return getattr(self, "_topic", None)

    @topic.setter
    def topic(self, value):
        self._topic = value

    # def get_topic(self) -> str:
    #     return self.topic

    def process(self, command=None, result: Result = None, mqtt_broker=None, device_info=None):
        log.info("Using output processor: MQTT, topic: %s", self.topic)
        log.debug("formatter: %s, result: %s, mqtt_broker: %s, device_info: %s", self.formatter, result, mqtt_broker, device_info)

        # exit if no data
        if result is None:
            log.debug("No result to output")
            return

        # Not sure that formatter and mqtt_broker are set, could use builder pattern to ensure they are set
        if self.formatter is None:
            log.error("No formatter supplied")
            raise RuntimeError("No formatter supplied")

        if mqtt_broker is None:
            log.error("No mqtt broker supplied")
            raise RuntimeError("No mqtt broker supplied")

        # build the messages...
        formatted_data = self.formatter.format(command=command, result=result, device_info=device_info)
        log.debug("mqtt.output msgs %s", formatted_data)

        # publish
        if isinstance(formatted_data, (str, bytes)):
            # simple payload, so publish as payload
            mqtt_broker.publish(topic=self.topic, payload=formatted_data)
        elif isinstance(formatted_data, list):
            # iterate list
            for item in formatted_data:
                if isinstance(item, (str, bytes)):
                    mqtt_broker.publish(topic=self.topic, payload=item)
                elif isinstance(item, dict) and 'topic' in item and 'payload' in item:
                    mqtt_broker.publish(topic=item['topic'], payload=item['payload'])
                elif isinstance(item, dict) and 'payload' in item:
                    mqtt_broker.publish(topic=self.topic, payload=item['payload'])
                else:
                    log.warning('Unknown mqtt data to publish, type: %s, data: %s', type(item), item)
        else:
            log.warning('Unknown mqtt data to publish, type: %s, data: %s', type(formatted_data), formatted_data)

    @classmethod
    def from_config(cls, output_config) -> "MQTT":
        """build object from config dict"""
        topic = output_config.get("topic", None)
        return cls(topic=topic)



""" reading.py """
from powermon.dto.reading_dto import ReadingDTO


class Reading:
    """ class to contain the raw reading, processed reading and reading definition for a particular reading """
    def __str__(self):
        return f"Reading: {self.data_name=}, {self.data_value=}, {self.data_unit=}, {self.is_valid=}"

    # def __init__(self, data_name: str, data_value: str, data_unit: str, device_class: str = None, state_class: str = None, icon: str = None) -> None:
    def __init__(self, raw_value, processed_value, definition) -> None:
        self.raw_value = raw_value
        self.processed_value = processed_value
        self.definition = definition
        self.is_valid = True

    def to_dto(self) -> ReadingDTO:
        """ convert Reading to a data transfer object """
        return ReadingDTO(data_name=self.data_name, data_value=self.data_value, data_unit=self.data_unit)

    @property
    def raw_value(self):
        """ the raw reading returned from the device - as returned """
        return getattr(self, "_raw_value", None)

    @raw_value.setter
    def raw_value(self, value):
        self._raw_value = value

    @property
    def processed_value(self):
        """ the reading after processing and conversion - the reading we want to see """
        return getattr(self, "_processed_value", None)

    @processed_value.setter
    def processed_value(self, value):
        # should take raw value and 'process it'
        self._processed_value = value

    @property
    def definition(self):
        """ the reading definition associated with this reading """
        return getattr(self, "_definition", None)

    @definition.setter
    def definition(self, value):
        self._definition = value

    # the below are / should be part of reading definition
    @property
    def data_name(self) -> str:
        return self.definition.description

    @property
    def data_unit(self) -> str:
        if self.definition.unit is None:
            return ""
        return self.definition.unit

    @property
    def data_value(self) -> str:
        return self.processed_value

    @property
    def icon(self) -> str | None:
        return self.definition.icon

    @property
    def device_class(self) -> str | None:
        return self.definition.device_class

    @property
    def state_class(self) -> str | None:
        return self.definition.state_class



""" powermon / ports / serialport.py """
# import serial
import asyncio
import logging

from bleak import BleakClient

from powermon.commands.command import Command, CommandType
from powermon.commands.result import Result
from powermon.dto.portDTO import PortDTO
from powermon.errors import PowermonWIP
from powermon.ports.abstractport import AbstractPort
from powermon.ports.porttype import PortType
from powermon.protocols import get_protocol_definition

log = logging.getLogger("BlePort")


class BlePort(AbstractPort):
    """ BLE port object """

    def __str__(self):
        return f"BlePort: {self.mac=}, protocol:{self.protocol}, {self.client=}, {self.error_message=}"

    @classmethod
    def from_config(cls, config=None):
        log.debug("building ble port. config:%s", config)
        mac = config.get("mac")
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(mac=mac, protocol=protocol)

    def __init__(self, mac, protocol) -> None:
        super().__init__(protocol=protocol)
        self.port_type = PortType.SERIAL
        self.mac = mac
        self.client = None
        self.is_protocol_supported()
        # self.error_message = None
        self.response_cache = {}

    def to_dto(self) -> PortDTO:
        dto = PortDTO(type="ble", mac=self.mac, protocol=self.protocol.to_dto())
        return dto

    def _notification_callback(self, handle, data):
        log.debug("%s %s %s", handle, repr(data), len(data))
        print(f"callback - {handle=}, {data=}")
        return
        # responses = []
        # command_code =90
        # if len(data) == 13:
        #     responses.append(data)
        # elif len(data) == 26:
        #     responses.append(data[0:13])
        #     responses.append(data[13:])
        # else:
        #     # self.logger.error(len(data), "bytes received, not 13 or 26, not implemented")
        #     pass

        # for response_bytes in responses:
        #     #command = response_bytes[2:3].hex()
        #     if self.response_cache["done"] is True:
        #         # self.logger.debug("skipping response for %s, done" % command)
        #         return
        #     self.response_cache["queue"].append(response_bytes[4:-1])
        #     if len(self.response_cache["queue"]) == self.response_cache["max_responses"]:
        #         self.response_cache["done"] = True
        #         self.response_cache["future"].set_result(self.response_cache["queue"])

    def disconnect_callback(self, client):
        """ callback for disconnection """
        print(f"disconnect callback, {client=}")
        self.client = None

    def is_connected(self):
        return self.client is not None and self.client.is_connected

    async def connect(self) -> int:
        log.debug("bleport connecting. mac:%s", self.mac)
        try:
            # build client object
            self.client = BleakClient(self.mac, disconnected_callback=self.disconnect_callback)
            # connect to client
            await self.client.connect()
            # 'turn on' notification characteristic
            # await self.client.start_notify(17, self._notification_callback)
            # write to 'XXX' charateristic
            # await self.client.write_gatt_char(48, bytearray(b""))
            log.debug(self.client)
        except Exception as e:
            # log.error("Incorrect configuration for serial port: %s", e)
            # self.error_message = str(e)
            print(e)
            self.client = None
            raise PowermonWIP("connect failed") from e
        
        return self.is_connected()

    async def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.client is not None:
            await self.client.disconnect()
        self.client = None

    async def send_and_receive(self, command: Command) -> Result:
        full_command = command.full_command
        command_code = 90
        self.response_cache = {"queue": [],
                                "future": asyncio.Future(),
                                "max_responses": 3,
                                "done": False}
        response_line = None
        log.debug("port: %s, full_command: %s", self.client, full_command)
        if not self.is_connected():
            raise RuntimeError("Ble port not open")
        # try:
        log.debug("Executing command via ble...")
        await self.client.write_gatt_char(15, full_command)
        log.debug("Waiting...")
        try:
            response_line = await asyncio.wait_for(self.response_cache["future"], 30)
        except asyncio.TimeoutError:
            log.warning("Timeout while waiting for %s response" % command)
            exit(1)
            return False
        print("got %s" % response_line)
        #return result
        # self.serial_port.reset_input_buffer()
        # self.serial_port.reset_output_buffer()
        # # Process i/o differently depending on command type
        # command_defn = command.command_definition
        # match command_defn.command_type:
        #     case CommandType.VICTRON_LISTEN:
        #         # this command type doesnt need to send a command, it just listens on the serial port
        #         _lines = 30
        #         log.debug("VictronCommandType.LISTEN s&r, listening for %i lines", _lines)
        log.debug("serial response was: %s", response_line)
        # response = self.get_protocol().check_response_and_trim(response_line)
        result = command.build_result(raw_response=response_line, protocol=self.protocol)
        return result
        # except Exception as e:
        #     log.warning("Serial read error: %s", e)
        #     result.error = True
        #     result.error_messages.append(f"Serial read error {e}")
        #     self.disconnect()
        #     return result



"""device.py"""
import logging


from powermon.commands.command import Command
from powermon.commands.result import Result
from powermon.dto.deviceDTO import DeviceDTO
from powermon.errors import CommandDefinitionMissing, ConfigError
from powermon.libs.mqttbroker import MqttBroker
from powermon.outputs.abstractoutput import AbstractOutput
from powermon.ports import from_config as port_from_config
from powermon.ports.abstractport import AbstractPort

# Set-up logger
log = logging.getLogger("Device")


class DeviceInfo:
    """ struct like class to contain info about the device """
    def __init__(self, name, device_id, model, manufacturer):
        self.name = name
        self.device_id = device_id
        self.model = model
        self.manufacturer = manufacturer


class Device:
    """
    A device is a port with a protocol
    also contains the name, model and id of the device
    """

    def __str__(self):
        return f"Device: {self.device_info.name}, {self.device_info.device_id=}, " + \
            f"{self.device_info.model=}, {self.device_info.manufacturer=}, " + \
            f"port: {self.port}, mqtt_broker: {self.mqtt_broker}, commands:{self.commands}"

    @classmethod
    def from_config(cls, config=None):
        """build the object from a config dict"""
        if not config:
            log.warning("No device definition in config. Check configFile argument?")
            return cls(name="unnamed")
        name = config.get("name", "unnamed_device")
        device_id = config.get("id", "1")  # device_id needs to be unique if there are two devices
        model = config.get("model")
        manufacturer = config.get("manufacturer")

        port = port_from_config(config.get("port"))

        # raise error if unable to configure port
        if not port:
            log.error("Invalid port config '%s' found", config)
            raise ConfigError(f"Invalid port config '{config}' found")

        return cls(name=name, device_id=device_id, model=model, manufacturer=manufacturer, port=port)

    def __init__(self, name: str, device_id: str = "", model: str = "", manufacturer: str = "", port: AbstractPort = None):
        self.device_info = DeviceInfo(name=name, device_id=device_id, model=model, manufacturer=manufacturer)
        self.port: AbstractPort = port
        self.commands: list[Command] = []
        self.mqtt_broker = None
        self.adhoc_commands: list = []

    @property
    def port(self) -> AbstractPort:
        """ the port associated with this device """
        return self._port

    @port.setter
    def port(self, value):
        log.debug("Setting port to: %s", value)
        self._port = value

    @property
    def mqtt_broker(self) -> MqttBroker:
        """ the mqtt_broker object """
        return self._mqtt_broker

    @mqtt_broker.setter
    def mqtt_broker(self, mqtt_broker):
        log.debug("Setting mqtt_broker to: %s", mqtt_broker)
        self._mqtt_broker = mqtt_broker
        # If we are setting an actual broker object also set callback for adhoc commands
        if isinstance(mqtt_broker, MqttBroker) and mqtt_broker.adhoc_topic is not None:
            log.info("Subscribing to adhoc_topic: %s", mqtt_broker.adhoc_topic)
            mqtt_broker.subscribe(topic=mqtt_broker.adhoc_topic, callback=self.adhoc_command_cb)

    def adhoc_command_cb(self, client, userdata, msg):
        """ callback for adhoc command messages """
        log.debug("received message on %s, with payload: %s", msg.topic, msg.payload)
        # build command object
        command_code = msg.payload.decode()
        adhoc_command = Command.from_code(command_code)
        adhoc_command.command_definition = self.port.protocol.get_command_definition(command_code)
        # add to adhoc queue
        self.adhoc_commands.append(adhoc_command)

    def add_command(self, command: Command) -> None:
        """add a command to the devices' list of commands"""
        if command is None:
            return
        # get command definition from protocol
        try:
            command.command_definition = self.port.protocol.get_command_definition(command.code)
        except CommandDefinitionMissing as ex:
            print(ex)
            return

        # append to commands list
        self.commands.append(command)
        log.debug("added command (%s), command list length: %i", command, len(self.commands))

    def to_dto(self) -> DeviceDTO:
        """convert the Device to a Data Transfer Object"""
        commands = []
        command: Command
        for command in self.commands:
            commands.append(command.to_dto())
        dto = DeviceDTO(device_id=self.device_info.device_id,
                        model=self.device_info.model,
                        manufacturer=self.device_info.manufacturer,
                        port=self.port.to_dto(),
                        commands=commands)
        return dto

    async def initialize(self):
        """Device initialization activities"""
        log.info("initializing device")

    async def finalize(self):
        """Device finalization activities"""
        log.info("finalizing device")
        # close connection on port
        await self.port.disconnect()

    async def run_adhoc_commands(self):
        """ check for any adhoc commands in the queue and run them """
        # check for any adhoc commands
        while len(self.adhoc_commands) > 0:
            # get the oldest adhoc command
            adhoc_command = self.adhoc_commands.pop(0)
            # run command
            log.info("Running adhoc command: %s", adhoc_command)
            try:
                # run command
                result: Result = await self.port.run_command(adhoc_command)
                log.info("Got result: %s", result)
            except Exception as exception:  # pylint: disable=W0718
                # specific errors need to incorporated into Result as part of the processing
                # so any exceptions at this stage will be truely unexpected
                log.error("Error decoding result: %s", exception)
                # result = Result(result_type=ResultType.ERROR, command_definition=None, \
                #   raw_response=b"Error decoding result", trimmed_response=b"Error decoding result")
                # result.error_messages.append(f"Error decoding result: {exception}")
                # result.error_messages.append(f"Exception Type: {exception.__class__.__name__}")
                # result.error_messages.append(f"Exception args: {exception.args}")
                raise exception
            # process result
            payload = str(result)  # FIXME: finish this
            # publish result
            print(payload)
            self.mqtt_broker.publish("powermon2/adhoc_commands_results", payload)  # FIXME: finish this

    async def run(self, force=False):
        """checks for commands to run and runs them"""
        # run any adhoc commands
        await self.run_adhoc_commands()

        # check for any commands in the queue
        if self.commands is None or len(self.commands) == 0:
            log.info("no commands in queue")
            return

        for command in self.commands:
            if force or command.is_due():
                log.info("Running command: %s", command)
                try:
                    # run command
                    result: Result = await self.port.run_command(command)
                    log.info("Got result: %s", result)
                except Exception as exception:  # pylint: disable=W0718
                    # specific errors need to incorporated into Result as part of the processing
                    # so any exceptions at this stage will be truely unexpected
                    log.error("Error decoding result: %s", exception)
                    # result = Result(result_type=ResultType.ERROR, command_definition=None, \
                    #   raw_response=b"Error decoding result", trimmed_response=b"Error decoding result")
                    # result.error_messages.append(f"Error decoding result: {exception}")
                    # result.error_messages.append(f"Exception Type: {exception.__class__.__name__}")
                    # result.error_messages.append(f"Exception args: {exception.args}")
                    raise exception
                # result.device_id = self.device_id

                # loop through each output and process result
                output: AbstractOutput
                for output in command.outputs:
                    log.debug("Using Output: %s", output)
                    output.process(command=command, result=result, mqtt_broker=self.mqtt_broker, device_info=self.device_info)
